import sympy
from sympy.parsing.sympy_parser import parse_expr
import numpy as np
import scipy.optimize
import matplotlib.pyplot as plt
# from numpy.random import default_rng

#### Wrapper functions ####
def init_general_test_point(eq, X, y, bitmask):
    parsed_eq = parse_expr(eq)
    symbols = parsed_eq.free_symbols
    symbol_list = [x.name for x in symbols]
    symbol_list.sort(key=lambda x: int(x[1:]))
    return TestPoints(eq, symbol_list, X, y, bitmask)

def get_test_point(obj, full_pt):
    ''' Takes a N-dim vector, return N-dim vector '''
    pt = full_pt[obj.bitmask]
    opt = obj.find_reference_point(pt)
    in_pt = np.copy(full_pt)
    in_pt[obj.bitmask] = opt
    return in_pt


#### UTILS ####
# rng = default_rng()
def relative_error(hypot, bench):
    return np.abs((hypot-bench)/bench)

def project_plane(normal, point, guess):
    normal /= np.linalg.norm(normal)
    return guess - np.dot(guess-point, normal)*normal

class TestPoints:
    def __init__(self, eq, symbols, X, y, bitmask, mode='general'):
        '''
        mode is in {'general', 'add', 'minus', 'times', 'divide'} 
        eq, symbols are None if mode is not general
        '''
        self.mode = mode
        if mode == 'general':
            self.general_init(eq, symbols, X, y, bitmask)
        else:
            eq_map = {
                'add': 'x+y',
                'minus': 'x-y',
                'times': 'x*y',
                'divide': 'x/y',
            }
            if mode in eq_map:
                self.general_init(eq_map[mode], ['x', 'y'], X, y, bitmask)
            else:
                raise Exception("Unknown mode "+mode)
    def general_init(self, eq, symbols, X, y, bitmask):
        self.eq = eq
        self.symp_eq = sympy.sympify(eq)
        self.symbols = [sympy.symbols(x) for x in symbols]
        self.X = X
        self.y = y
        self.bitmask = bitmask
        self.symp_grads = [sympy.diff(self.symp_eq, x) for x in symbols]
        self.lambda_eq = sympy.lambdify(self.symbols, self.symp_eq)
        self.lambda_grads = [sympy.lambdify(self.symbols, x) for x in self.symp_grads]
        self.low_range = np.percentile(self.X[:, self.bitmask], 0, axis=0)
        self.high_range = np.percentile(self.X[:, self.bitmask], 100, axis=0)
        
        self.init_median_projection()
        self.init_scatter()

        self.log = []

    def init_median_projection(self):
        self.median_point = np.median(self.X[:, self.bitmask], axis=0)

    def init_scatter(self):
        self.hval = self.lambda_eq(*[self.X[:, i] for i in range(np.sum(self.bitmask))])
        self.hindices = np.argsort(self.hval)

    def evaluate_gradients(self, pt):
        r = np.array([f(*pt) for f in self.lambda_grads]).astype(float)
        return r

    def find_initial_guess_median_projection(self, pt):
        num_grads = self.evaluate_gradients(pt)
        num_grads /= np.linalg.norm(num_grads)
        return [project_plane(num_grads, pt, self.median_point)]
    

    def find_initial_guess_scatter(self, pt, low=2, high=3):
        guess = self.find_initial_guess_median_projection(pt)
        target_hval = self.lambda_eq(*pt)
        target_index = np.searchsorted(self.hval, target_hval, sorter=self.hindices)
        candidates = list(range(max(0, target_index-low), min(self.X.shape[0], target_index+high)))
        results = [self.X[self.hindices[guess], self.bitmask] for guess in candidates]
        return results
    
    def find_initial_guess_random(self, pt, num=2):
        return [self.X[t, self.bitmask] for t in np.random.randint(0, self.X.shape[0], num)]

    def optimize_fmin(self, guess, target):
        FTOL = 1e-4
        MAXITER = 200
        result = scipy.optimize.fmin(
            lambda x: np.abs(self.lambda_eq(*x) - target),
            guess, maxiter=MAXITER, ftol=FTOL, full_output=True, disp=False)
        if result[4] == 0:
            return result[0]
        else:
            return None
    def optimize_bfgs(self, guess, target):
        MAXITER = 20
        res = scipy.optimize.minimize(lambda x: .5 * (self.lambda_eq(*x) - target)**2,
                                    guess, method='BFGS', 
                                    jac=lambda x: self.evaluate_gradients(x) * (self.lambda_eq(*x) - target), 
                                    options={'maxiter':MAXITER, 'disp': False})
        if res.success:
            return res.x
        return None
    def optimize_basic(self, mode, guess, target):
        if mode == 'add':
            return project_plane(np.array([1., 1.]), np.array([target, 0]), guess)
        elif mode == 'minus':
            return project_plane(np.array([1., -1.]), np.array([target, 0]), guess)
        elif mode == 'divide':
            return project_plane(np.array([1., -target]), np.array([0, 0]), guess)
        elif mode == 'times':
            a, b = guess
            if target == 0:
                return np.array([a, 0]) if abs(a) >= abs(b) else np.array([0, b])
            else:
                A, B, C  = a, b**2 - a**2, -b*target
                x = (-B + (1 if target > 0 else -1) * (B**2 - 4*A*C)**0.5)/2/A
                return np.array([x, target/x])

    def in_domain(self, pt):
        return np.all(self.low_range <= pt) and np.all(pt <= self.high_range)
    def find_reference_point(self, pt):
        guesses = self.find_initial_guess_scatter(pt) + self.find_initial_guess_median_projection(pt) + self.find_initial_guess_random(pt)
        target_hval = self.lambda_eq(*pt)
        results = []
        for guess in guesses:
            if self.mode == 'general':
                result = self.optimize_bfgs(guess, target_hval)
            else:
                result = self.optimize_basic(self.mode, guess, target_hval)
            if result is not None and self.in_domain(result):
                results.append(result)
        return max(results, key=lambda x: np.linalg.norm(pt - x), default=None)
        
    
    def analyze_reference_point(self, pt, opt, disp):
        reference_point_rel_error = relative_error(self.lambda_eq(*opt), self.lambda_eq(*pt))
        reference_point_distance = np.linalg.norm(opt - pt)
        max_distance = np.linalg.norm(self.high_range - self.low_range)
        reference_point_rel_distance = reference_point_distance / max_distance
        self.log.append((reference_point_rel_error, reference_point_rel_distance))
#        if disp:
#            print(f'{pt} : found {opt}, err: {reference_point_rel_error}, distance: {reference_point_rel_distance}')

    def score_pt(self, model, full_pt, disp=False, log=False):
        pt = full_pt[self.bitmask]
        opt = self.find_reference_point(pt)
        if opt is None:
            return None
        if log:
            self.analyze_reference_point(pt, opt, disp)
        in_pt = np.copy(full_pt)
        in_pt[self.bitmask] = opt
        return relative_error(model(in_pt[np.newaxis, :]), model(full_pt[np.newaxis, :]))



import numpy as np
from .S_snap import bestApproximation
from .S_get_number_DL_snapped import get_number_DL_snapped
from sympy.parsing.sympy_parser import parse_expr
from sympy import preorder_traversal, count_ops


def get_expr_complexity(expr):
    expr = parse_expr(expr)
    compl = 0

    is_atomic_number = lambda expr: expr.is_Atom and expr.is_number
    numbers_expr = [subexpression for subexpression in preorder_traversal(expr) if is_atomic_number(subexpression)]

    for j in numbers_expr:
        try:
            compl = compl + get_number_DL_snapped(float(j))
        except:
            compl = compl + 1000000

    n_variables = len(expr.free_symbols)
    n_operations = len(count_ops(expr,visual=True).free_symbols)

    if n_operations!=0 or n_variables!=0:
        compl = compl + (n_variables+n_operations)*np.log2((n_variables+n_operations))

    return compl



import numpy as np
import pandas as pd
from scipy.sparse.linalg import lsqr
from scipy.linalg import *
from sympy import Matrix
from sympy import symbols, Add, Mul, S
from .getPowers import getPowers

def dimensional_analysis(input,output,units):
    M = units[input[0]]
    for i in range(1,len(input)):
        M = np.c_[M, units[input[i]]]
    if len(input)==1:
        M = np.array(M)
        M = np.reshape(M,(len(M),1))
    params = getPowers(M,units[output])
    M = Matrix(M)
    B = M.nullspace()
    return (params, B)

# load the data from a file
def load_data(pathdir, filename):
    n_variables = np.loadtxt(pathdir+filename, dtype='str').shape[1]-1
    variables = np.loadtxt(pathdir+filename, usecols=(0,))
    for i in range(1,n_variables):
        v = np.loadtxt(pathdir+filename, usecols=(i,))
        variables = np.column_stack((variables,v))
    f_dependent = np.loadtxt(pathdir+filename, usecols=(n_variables,))
    return(variables.T,f_dependent)

def dimensionalAnalysis(pathdir, filename, eq_symbols):
    file = pd.read_excel("units.xlsx")

    units = {}
    for i in range(len(file["Variable"])):
        val = [file["m"][i],file["s"][i],file["kg"][i],file["T"][i],file["V"][i],file["cd"][i]]
        val = np.array(val)
        units[file["Variable"][i]] = val

    dependent_var = eq_symbols[-1]

    file_sym = open(filename + "_dim_red_variables.txt" ,"w")
    file_sym.write(filename)
    file_sym.write(", ")

    # load the data corresponding to the first line (from mystery_world)
    varibs = load_data(pathdir,filename)[0]
    deps = load_data(pathdir,filename)[1]

    # get the data in symbolic form and associate the corresponding values to it
    input = []
    for i in range(len(eq_symbols)-1):
        input = input + [eq_symbols[i]]
        vars()[eq_symbols[i]] = varibs[i]
    output = dependent_var

    # Check if all the independent variables are dimensionless
    ok = 0
    for j in range(len(input)):
        if(units[input[j]].any()):
            ok=1

    if ok==0:
        dimless_data = load_data(pathdir, filename)[0].T
        dimless_dep = load_data(pathdir, filename)[1]
        if dimless_data.ndim==1:
            dimless_data = np.reshape(dimless_data,(1,len(dimless_data)))
            dimless_data = dimless_data.T
        np.savetxt(pathdir + filename + "_dim_red", dimless_data)
        file_sym.write(", ")
        for j in range(len(input)):
            file_sym.write(str(input[j]))
            file_sym.write(", ")
        file_sym.write("\n")
    else:
        # get the symbolic form of the solved part
        solved_powers = dimensional_analysis(input,output,units)[0]
        input_sym = symbols(input)
        sol = symbols("sol")
        sol = 1
        for i in range(len(input_sym)):
            sol = sol*input_sym[i]**np.round(solved_powers[i],2)
        file_sym.write(str(sol))
        file_sym.write(", ")

        # get the symbolic form of the unsolved part
        unsolved_powers = dimensional_analysis(input,output,units)[1]

        #print(unsolved_powers,unsolved_powers[0])
        uns = symbols("uns")
        unsolved = []
        for i in range(len(unsolved_powers)):
            uns = 1
            for j in range(len(unsolved_powers[i])):
                uns = uns*input_sym[j]**unsolved_powers[i][j]
            file_sym.write(str(uns))
            file_sym.write(", ")
            unsolved = unsolved + [uns]
        file_sym.write("\n")

        # get the discovered part of the function
        func = 1
        for j in range(len(input)):
            func = func * vars()[input[j]]**dimensional_analysis(input,output,units)[0][j]
        func = np.array(func)

        # get the new variables needed
        new_vars = []
        for i in range(len(dimensional_analysis(input,output,units)[1])):
            nv = 1
            for j in range(len(input)):
                nv = nv*vars()[input[j]]**dimensional_analysis(input,output,units)[1][i][j]
            new_vars = new_vars + [nv]

        new_vars = np.array(new_vars)
        new_dependent = deps/func

        if new_vars.size==0:
            np.savetxt(pathdir + filename + "_dim_red", new_dependent)

        # save this to file
        all_variables = np.vstack((new_vars, new_dependent)).T
        np.savetxt(pathdir + filename + "_dim_red", all_variables)

    file_sym.close()


#print(dimensionalAnalysis("../_noise_data/", "119_1.24.6", ["m","omega","omega_0","x","E_n"]))




# The following are snap functions for finding a best approximated integer or rational number for a real number:

import numpy as np
from sympy import Rational

def bestApproximation(x,imax):
    # The input is a numpy parameter vector p.
    # The output is an integer specifying which parameter to change,
    # and a float specifying the new value.
    def float2contfrac(x,nmax):
        x = float(x)
        c = [np.floor(x)];
        y = x - np.floor(x)
        k = 0
        while np.abs(y)!=0 and k<nmax:
            y = 1 / float(y)
            i = np.floor(y)
            c.append(i)
            y = y - i
            k = k + 1
        return c
    
    def contfrac2frac(seq):
        ''' Convert the simple continued fraction in `seq`
            into a fraction, num / den
            '''
        num, den = 1, 0
        for u in reversed(seq):
            num, den = den + num*u, num
        return num, den
    
    def contFracRationalApproximations(c):
        return np.array(list(contfrac2frac(c[:i+1]) for i in range(len(c))))
    
    def contFracApproximations(c):
        q = contFracRationalApproximations(c)
        return q[:,0] / float(q[:,1])
    
    def truncateContFrac(q,imax):
        k = 0
        while k < len(q) and np.maximum(np.abs(q[k,0]), q[k,1]) <= imax:
            k = k + 1
        return q[:k]
    
    def pval(p):
        p = p.astype(float)
        return 1 - np.exp(-p ** 0.87 / 0.36)
    
    xsign = np.sign(x)
    q = truncateContFrac(contFracRationalApproximations(float2contfrac(abs(x),20)),imax)
    
    if len(q) > 0:
        p = np.abs(q[:,0] / q[:,1] - abs(x)).astype(float) * (1 + np.abs(q[:,0])) * q[:,1]
        p = pval(p)
        i = np.argmin(p)
        return (xsign * q[i,0] / float(q[i,1]), xsign* q[i,0], q[i,1], p[i])
    else:
        return (None, 0, 0, 1)

def integerSnap(p, top=1):
    p = np.array(p)
    metric = np.abs(p - np.round(p.astype(np.double)))
    chosen = np.argsort(metric)[:top]
    return dict(list(zip(chosen, np.round(p.astype(np.double))[chosen])))


def zeroSnap(p, top=1):
    p = np.array(p)
    metric = np.abs(p)
    chosen = np.argsort(metric)[:top]
    return dict(list(zip(chosen, np.zeros(len(chosen)))))


def rationalSnap(p, top=1):
    """Snap to nearest rational number using continued fraction."""
    p = np.array(p)
    snaps = np.array(list(bestApproximation(x,10) for x in p))
    chosen = np.argsort(snaps[:, 3])[:top]    
    d = dict(list(zip(chosen, snaps[chosen, 1:3])))
    d = {k:  f"{val[0]}/{val[1]}" for k,val in d.items()}
    
    return d



